#define _XOPEN_SOURCE 1
#define _XOPEN_SOURCE_EXTENDED 1
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <stddef.h>
#include <stdarg.h>

static char const prog[]="svclean-shutdown";

static int status=0;
static void setstatus(void) {
  int const es=errstat(errno);
  if (status==0 || es==100) status=es;
}

static void printerr(char const* s0, char const* s1, char const* s2,
                     char const* s3, char const* s4) {
  int const error=errno;
  buffer_puts(buffer_2, prog);
  buffer_puts(buffer_2, err_);
  buffer_puts(buffer_2, err_unable);
  buffer_puts(buffer_2, s0);
  if (s1!=null) {
    buffer_puts(buffer_2, s1);
    if (s2!=null) {
      buffer_puts(buffer_2, s2);
      if (s3!=null) {
        buffer_puts(buffer_2, s3);
        if (s4!=null)
          buffer_puts(buffer_2, s4);
      }
    }
  }
  buffer_puts(buffer_2, err_);
  buffer_puts(buffer_2, error_str(error));
  buffer_puts(buffer_2, "\n");
  buffer_flush(buffer_2);
}

static void printerr_die(char const* s0, char const* s1, char const* s2,
                         char const* s3, char const* s4) {
  setstatus();
  printerr(s0, s1, s2, s3, s4);
  _exit(status);
}

static char const svclean[]="/" SVCLEAN;
static char const shutdown[]="/" SVCLEAN SVCLEAN_SHUTDOWN;
static char const control[]="/supervise/control";

static int servicesdir;
static char const* servicesdirname;
static struct taia one;
static struct taia six;

typedef struct srvlist srvlist;
struct srvlist {
  struct taia wake_at;
  srvlist* next;
  struct taia* sleep_time;
  char const* svc_command;
  unsigned long len;
  unsigned char next_op;
  char name[1];
};
#define srvlist_size(len) (offsetof(srvlist, name)+(len))

typedef void srvop_t(srvlist*);
static srvop_t op_create_shutdown;
static srvop_t op_sleep;
static srvop_t op_svc;
static srvop_t* const ops[]={
  op_create_shutdown, op_sleep, op_svc, op_sleep,
  op_create_shutdown, op_sleep, op_svc, op_sleep,
  op_svc
};
static unsigned int const ops_done=(sizeof ops)/(sizeof ops[0]);

static void op_create_shutdown(srvlist* service) {
  if (fchdir(servicesdir)!=0)
    printerr_die(err_chdirto, servicesdirname, null, null, null);
  if (chdir(service->name)!=0) {
    if (errno!=ENOENT) {
      setstatus();
      printerr(err_chdirto, servicesdirname, "/", service->name, svclean);
    }
    service->next_op=ops_done;
    return;
  }

  if (mkdir(svclean+1, 0755)!=0 && errno!=EEXIST) {
    setstatus();
    printerr(err_create, servicesdirname, "/", service->name, svclean);
    service->next_op=ops_done;
    return;
  }
  {
    int const fd=open(shutdown+1, O_WRONLY|O_CREAT, 0644);
    if (fd<0) {
      setstatus();
      printerr(err_create, servicesdirname, "/", service->name, shutdown);
      service->next_op=ops_done;
      return;
    }
    close(fd);
  }

  taia_now(&service->wake_at);
  taia_add(&service->wake_at, &service->wake_at, &one);
  service->svc_command="dx";
  service->next_op++;
}

static void op_sleep(srvlist* service) {
  struct taia now;
  taia_now(&now);
  if (taia_less(&now, &service->wake_at)) return;
  service->next_op++;
}

static void op_svc(srvlist* service) {
  int fd;

  if (fchdir(servicesdir)!=0)
    printerr_die(err_chdirto, servicesdirname, null, null, null);
  if (chdir(service->name)!=0) {
    setstatus();
    printerr(err_chdirto, servicesdirname, "/", service->name, svclean);
    service->next_op=ops_done;
    return;
  }

  fd=open(control+1, O_WRONLY|nonblock);
  if (fd<0) {
    if (errno==ENOENT || errno==ENXIO) {
      service->next_op++;
      if (service->next_op!=ops_done) {
        taia_now(&service->wake_at);
        taia_add(&service->wake_at, &service->wake_at, service->sleep_time);
        service->sleep_time=&six;
        service->name[service->len]='/';
      }
      return;
    }
    setstatus();
    printerr(err_open, servicesdirname, "/", service->name, control);
    service->next_op=ops_done;
    return;
  }

  if (service->svc_command!=null) {
    buffer buf;
    int const flags=fcntl(fd, F_GETFL, 0);
    char buf_space[2];

    if (flags==-1) {
      setstatus();
      printerr(err_getflags, null, null, null, null);
      service->next_op=ops_done;
      return;
    }

    if (fcntl(fd, F_SETFL, flags&~nonblock)==-1) {
      setstatus();
      printerr(err_setflags, null, null, null, null);
      service->next_op=ops_done;
      return;
    }

    buffer_init(&buf, buffer_unixwrite, fd, buf_space, sizeof buf_space);
    if (buffer_putsflush(&buf, service->svc_command)!=0) {
      setstatus();
      printerr("write commands to ", servicesdirname, "/", service->name,
               control);
      close(fd);
      service->next_op=ops_done;
      return;
    }
  }

  close(fd);

  service->svc_command=(service->name[service->len]=='\0'? null: "x");
}

int main(int argc, char** argv) {
  DIR* dirh;
  srvlist* services, * service, ** next;
  static char const log[]="/log";
  struct taia zero;

  if (argc!=2) {
    buffer_puts(buffer_2, prog);
    buffer_puts(buffer_2, usage);
    buffer_puts(buffer_2, prog);
    buffer_puts(buffer_2, " /service\n");
    buffer_flush(buffer_2);
    _exit(100);
  }

  servicesdirname=argv[1];
  servicesdir=open(servicesdirname, O_RDONLY);
  if (servicesdir<0)
    printerr_die(err_open, servicesdirname, null, null, null);
  dirh=opendir(servicesdirname);
  if (dirh==null)
    printerr_die(err_open, servicesdirname, null, null, null);

  services=null;
  next=&services;
  for (;;) {
    struct dirent const* de;
    unsigned long namelen, loglen, size;

    errno=0;
    de=readdir(dirh);
    if (de==null) {
      if (errno==0) break;
      printerr_die("read ", servicesdirname, null, null, null);
    }

    if (de->d_name[0]=='.') continue;

    namelen=str_len(de->d_name);
    loglen=namelen+sizeof log;
    size=srvlist_size(loglen);
    if (loglen<sizeof log || size<loglen || size!=(unsigned int)size) {
      /* arithmetic overflow; probably impossible */
      buffer_puts(buffer_2, prog);
      buffer_puts(buffer_2, err_);
      buffer_puts(buffer_2, "name too long: ");
      buffer_puts(buffer_2, de->d_name);
      buffer_puts(buffer_2, "\n");
      buffer_flush(buffer_2);
      _exit(100);
    }

    service=(srvlist*)alloc((unsigned int)size);
    if (service==null)
      printerr_die(err_alloc, null, null, null, null);

    service->next=null;
    service->len=namelen;
    byte_copy(service->name, namelen, de->d_name);

    *next=service;
    next=&service->next;
  }
  closedir(dirh);

  if (services==null) _exit(status);

  taia_uint(&zero, 0u);
  taia_uint(&one, 1u);
  taia_uint(&six, 6u);

  for (service=services; service!=null; service=service->next) {
    byte_copy(service->name+service->len, sizeof log, log);
    service->name[service->len]='\0';
    service->next_op=0;
    service->sleep_time=&one;
  }

  for (;;) {
    /* are all services completely shut down? */
    unsigned int all_done=1;
    /* are any either polling or ready to do something? */
    unsigned int any_not_sleeping=0;
    /* are none waiting for a certain interval after some event? */
    unsigned int none_sleeping=1;
    /* are all either waiting or polling? */
    unsigned int do_sleep=1;
    /* when will we next have something to do? */
    struct taia wake_at;

    for (service=services; service!=null; service=service->next) {
      unsigned int const last_op=service->next_op;
      if (last_op==ops_done) continue;
      ops[service->next_op](service);
      if (service->next_op==ops_done) continue;
      all_done=0;
      if (ops[service->next_op]==op_sleep) {
        if (none_sleeping || taia_less(&service->wake_at, &wake_at))
          taia_add(&wake_at, &service->wake_at, &zero);
        none_sleeping=0;
      } else {
        any_not_sleeping=1;
        if (service->next_op!=last_op) do_sleep=0;
      }
    }

    if (all_done) _exit(status);

    if (do_sleep) {
      iopause_fd iop;
      struct taia now, now_plus_one;
      taia_now(&now);
      taia_add(&now_plus_one, &now, &one);
      if (none_sleeping ||
          (any_not_sleeping && taia_less(&now_plus_one, &wake_at)))
        /* at least one is polling; we won't sleep too long */
        iopause(&iop, 0, &now_plus_one, &now);
      else
        iopause(&iop, 0, &wake_at, &now);
    }
  }
}
